use regex::Regex;
use std::collections::{HashMap, VecDeque};

struct Monke {
    operation: Box<dyn Fn(usize) -> usize>,
    test: usize,
    result_recipient: (usize, usize),
    items: VecDeque<usize>,
}

enum Star {
    Silver,
    Gold,
}

fn main() {
    let mut monkes = HashMap::new();

    let s = std::fs::read_to_string("/dev/stdin").unwrap();
    let thing = s.split("\n\n");

    let mut monke_magic_number = 1;
    for monke_str in thing {
        let (monke, id) = monke_parser(monke_str);
        monke_magic_number *= monke.test;
        monkes.insert(id, monke);
    }

    let mut inspections_silver: HashMap<usize, usize> = HashMap::new();
    let mut inspections_gold = inspections_silver.clone();

    monke_processor(&mut monkes, &mut inspections_silver, Star::Silver, 0);
    monke_processor(
        &mut monkes,
        &mut inspections_gold,
        Star::Gold,
        monke_magic_number,
    );

    println!(
        "P1: {}, P2: {}",
        calc_monke_business(&inspections_silver),
        calc_monke_business(&inspections_gold)
    );
}

fn calc_monke_business(inspections: &HashMap<usize, usize>) -> usize {
    let mut ins = inspections.iter().map(|v| v.1).collect::<Vec<_>>();
    ins.sort();
    ins.pop().unwrap() * ins.pop().unwrap()
}

fn monke_processor(
    monkes: &mut HashMap<usize, Monke>,
    inspections: &mut HashMap<usize, usize>,
    part: Star,
    monke_magic_number: usize,
) {
    let num_turns = if matches!(part, Star::Gold) {
        10_000
    } else {
        20
    };

    for _ in 0..num_turns {
        for id in 0..monkes.len() {
            let monke = monkes.get_mut(&id).unwrap();
            let mut pass_queue = Vec::new();

            let mut inspections_for_monke = *inspections.get(&id).unwrap_or(&0);

            while let Some(mut worry_level) = monke.items.pop_front() {
                inspections_for_monke += 1;
                worry_level = (monke.operation)(worry_level);

                match part {
                    Star::Silver => worry_level /= 3,
                    Star::Gold => worry_level %= monke_magic_number,
                }

                let passes_test = worry_level % monke.test == 0;
                let recipient_id = if passes_test {
                    monke.result_recipient.0
                } else {
                    monke.result_recipient.1
                };
                pass_queue.push((recipient_id, worry_level));
            }

            inspections.insert(id, inspections_for_monke);

            for (recipient_id, val) in pass_queue {
                let recipient_monke = monkes.get_mut(&recipient_id).unwrap();
                recipient_monke.items.push_back(val);
            }
        }
    }
}

fn monke_parser(monke_str: &str) -> (Monke, usize) {
    let mut id = 0;
    let mut items = VecDeque::default();
    let mut operation: Box<dyn Fn(usize) -> usize>;
    let mut test = 0;
    let mut result_recipient = (0, 0);
    operation = Box::new(|n| n);

    for mut line in monke_str.lines() {
        line = line.trim();
        if line.starts_with("Monkey") {
            id = Regex::new(r"\d+")
                .unwrap()
                .find(line)
                .unwrap()
                .as_str()
                .parse::<usize>()
                .unwrap();
        }
        if line.starts_with("Starting") {
            items = Regex::new(r"\d+(\d+, )?")
                .unwrap()
                .find_iter(line)
                .map(|item| item.as_str().parse::<usize>().unwrap())
                .collect::<VecDeque<usize>>();
        }
        if line.starts_with("Operation") {
            let op = Regex::new(r"[*+/-].+")
                .unwrap()
                .find(line)
                .unwrap()
                .as_str()
                .split_once(' ')
                .unwrap();

            operation = match op.1 {
                "old" => match op.0 {
                    "+" => Box::new(|n| n + n),
                    "*" => Box::new(|n| n * n),
                    _ => panic!("https://music.portal2sounds.com/2627"),
                },
                _ => {
                    let num = op.1.parse::<usize>().unwrap();
                    match op.0 {
                        "+" => Box::new(move |n| n + num),
                        "*" => Box::new(move |n| n * num),
                        _ => panic!("https://music.portal2sounds.com/2627"),
                    }
                }
            };
        }
        if line.starts_with("Test") {
            test = Regex::new(r"[0-9]+")
                .unwrap()
                .find(line)
                .unwrap()
                .as_str()
                .parse::<usize>()
                .unwrap();
        }
        if line.starts_with("If true") {
            result_recipient.0 = Regex::new(r"[0-9]")
                .unwrap()
                .find(line)
                .unwrap()
                .as_str()
                .parse::<usize>()
                .unwrap();
        }
        if line.starts_with("If false") {
            result_recipient.1 = Regex::new(r"[0-9]")
                .unwrap()
                .find(line)
                .unwrap()
                .as_str()
                .parse::<usize>()
                .unwrap();
        }
    }

    (
        Monke {
            items,
            operation,
            test,
            result_recipient,
        },
        id,
    )
}
